#!/usr/bin/perl -w
#############################################################################
# alert.pl
# Written by: Brandon Zehm <caspian@dotconf.net>
#
# License:
#
# logdog.pl (hereafter referred to as "program") is free software;
#   you can redistribute it and/or modify it under the terms of the GNU General
#   Public License as published by the Free Software Foundation; either version
#   2 of the License, or (at your option) any later version.
# Note that when redistributing modified versions of this source code, you
#   must ensure that this disclaimer and the above coder's names are included
#   VERBATIM in the modified code.
#
# Disclaimer:
# This program is provided with no warranty of any kind, either expressed or
#   implied.  It is the responsibility of the user (you) to fully research and
#   comprehend the usage of this program.  As with any tool, it can be misused,
#   either intentionally (you're a vandal) or unintentionally (you're a moron).
#   THE AUTHOR(S) IS(ARE) NOT RESPONSIBLE FOR ANYTHING YOU DO WITH THIS PROGRAM
#   or anything that happens because of your use (or misuse) of this program,
#   including but not limited to anything you, your lawyers, or anyone else
#   can dream up.  And now, a relevant quote directly from the GPL:
#
#                           NO WARRANTY
#
#  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
# FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
# OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
# PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
# OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
# TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
# PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
# REPAIR OR CORRECTION.
#
#   ---
#
# Whee, that was fun, wasn't it?  Now let's all get together and think happy
#   thoughts - and remember, the same class of people who made all the above
#   legal spaghetti necessary are the same ones who are stripping your rights
#   away with DVD CCA, DMCA, stupid software patents, and retarded legal
#   challenges to everything we've come to hold dear about our internet.
#   So enjoy your dwindling "freedom" while you can, because 1984 is coming
#   sooner than you think.  :[
#
#############################################################################
#
# Purpose: To create a program that can run an alert command, monitor the
#          status of that commands output, and if it fails run a different
#          command.  It should have a very flexable configuration file that
#          allows creation of "alert chains" - chains of alerts, each with 
#          their own alert command, fallback-command, and other advanced 
#          options.  It must have a method of passing an arbatrary message
#          from the command line into any of the alert commands defined in
#          the configuration file.
#          
#
# Layout:  
#          Alert Manager is designed to be run on Linux or any other OS that 
#          properly supports posix threads.  Win32 is not supported by Alert
#          Manager.  The basic program flow looks like this: 
#            * Start
#            * Process command line
#            * Read "General Settings" from the configuration file
#            * Read the information about the alert we are about to run from the 
#              configuration file
#            * Enter a while (1) loop until the alert is completed successfully -
#              if a failure-alert command needs to run it is started in a separate
#              thread.  If the current alert fails the fallback command is loaded
#              and it is run (in the same while loop).  The while loop doesn't
#              exit unless some command completes successfully.
#              
#          
# Features:
#       - Doesn't keep the log file open longer than it needs to - so multiple
#         instances of alert.pl hopefully won't overwrite eachother's log files.
#       - Supports a *very* forgiving configuration file that allows pound,
#         double-slash, and /* */ C-style comments.
#       - A command line switch to check the validity of the configuration file.
#       - A method to aviod infinite looping (failing?) alert commands.
#       
#       
#############################################################################
use strict;
use IO::Socket;



## Global Variable(s)
my %conf = (
    "programName"          => $0,                                ## The name of this program
    "version"              => '1.0-RC2',                         ## The version of this program
    "authorName"           => 'Brandon Zehm',                    ## Author's Name
    "authorEmail"          => 'caspian@dotconf.net',             ## Author's Email Address
    "debug"                => 0,                                 ## Default debug level
    "configurationFile"    => '/etc/alertManager.conf',          ## Default configuration file
    
    ## The following entries are used later in the program
    "mode"                 => 'initialization',                  ## State of the program
    "logging"              => '',                                ## If this is true the printmsg function prints to the log file
    "retriesCounter"       => 0,
    "alertFailures"        => 0,                                 ## This keeps track of the total number of alert failures to avoid infinite loops
    
    ## The folling entries get their values from the configuration file
    "logFile"              => '',
    "maxAlertFailures"     => '',
    "message"              => '',
    
);
$conf{'programName'} =~ s/(.)*[\/,\\]//;                         ## Remove path from filename
$0 = "[$conf{'programName'}]";


## Variables populated by the configuration file
my %alert = ();
my @userVars = ();


#############################
##
##      MAIN PROGRAM
##
#############################

## Initialize
initialize();

## Process the command line
processCommandLine();

## Validate the configuration file if that's our mode
if ($conf{'mode'} eq "validate") {
    if (validateConfigurationFile($conf{'configurationFile'})) {
        printmsg($conf{'error'}, 0) if ($conf{'error'});
        quit($conf{'error'}, 1);
    }
    else {
        quit("Syntax OK.", 0);
    }
}

## Print a list of available commands if that's our mode
if ($conf{'mode'} eq "list") {            
    my @list = returnConfigurationFileSections($conf{'configurationFile'});
    if (scalar(@list)) {
        print "Available alert definitions:\n";
        foreach (grep(!/General Settings/, @list)) { print "    $_\n"; }
        print "\n";
        quit("",0);
    }
    else {
        quit("There are no alert definitions in the configuration file!", 1);
    }
}

## All other initialization stuff
handleHUP();

##  Drop permissions
$> = $conf{'UID'} if ($conf{'UID'});
$) = $conf{'GID'} if ($conf{'GID'});









 #######################
##                     ##
##  MAIN PROGRAM LOOP  ##
##                     ##
 #######################

## Loop until some command executes successfully (at which point we simply quit())
while (1) {
    
    $0 = "$0 -> $conf{'alert'}";
    
    %alert = ();
    my $textOutput = "";
    my $exitStatus = "";
    
    ## Prevent infinite loops of failing commands
    if ( ($conf{'maxAlertFailures'}) && ($conf{'alertFailures'} >= $conf{'maxAlertFailures'}) ) {
        quit("OS-ALERT - The maximum number of consecutive alert failures was reached while running alert [$conf{'alert'}].  Your message may be lost!  The original message (if any) was [$conf{'message'}]", 1);
    }
    
    
    #################################################
    ##  If the current alert is a special command  ##
    #################################################
    
    ## Log the message
    if ($conf{'alert'} =~ s/^LOG\s+//) {
        printmsg("OS-DEBUG - Special alert [LOG] - Writing the message to the file [$conf{'alert'}]", 1);
        ## Make sure our file exists, and if the file doesn't exist then create it
        if ( ! -f $conf{'alert'} ) {
            printmsg("NOTICE: The file [$conf{'alert'}] does not exist.  Creating it now with mode [0600].", 0);
            chmod (0600, $conf{'alert'}) unless ($^O =~ /win/i);
        }
        ## Now open the file and attach it to a filehandle
        if (! open (FILE,">>$conf{'alert'}")) {
            printmsg("OS-ERROR - Opening the file [$conf{'alert'}] for appending returned the error: $!", 0);
        }
        else {
            ## Now print the current message to the file
            print FILE "$conf{'message'}\n";
            close FILE;
            quit("OS-SUCCESS - The alert command [LOG $conf{'alert'}] completed successfully", 0);
        }
        quit("OS-ERROR - There was an error while trying to write the message [$conf{'message'}] to the file [$conf{'alert'}]", 1);
    }
    
    ## Syslog the message
    elsif ($conf{'alert'} =~ /^SYSLOG$/) {
        if ( syslog("/dev/log", 6, 1, $conf{'message'}) ) {
            quit("I was instructed to syslog a message, but syslog() returned the error [$conf{'error'}].  The message was [$conf{'message'}]", 1); 
        }
        quit("OS-SUCCESS - The alert command [SYSLOG] completed successfully", 0);
    }
    
    
    
    
    
    ##############################
    ##  Load the current alert  ##
    ##############################
    readConfigurationFile($conf{'configurationFile'}, $conf{'alert'}, \%alert) && do {
        quit("$conf{'error'}", 1);
    };
    printmsg ("Alert Configuration: ", 3); foreach my $key (keys(%alert))    { printmsg ("Key: $key =>\t Value: $alert{$key}", 3) if ($key and $alert{$key}); }
    
    
    
    
    
    ##############################
    ##  Run the system command  ##
    ##############################
    
    ## Prepare the command to be executed
    $alert{'commandTmp'} = $alert{'command'};
    $alert{'commandTmp'} =~ s/MESSAGE/$conf{'message'}/g if ($conf{'message'});
    
    ## Evaluate user defined variables in alert commands
    foreach (@userVars) {
        printmsg("Preparing alert command for running by replaceing user var [$_] with [" . $conf{"VAR:$_"} . "]", 2);
        $alert{'commandTmp'} =~ s/$_/$conf{"VAR\:$_"}/g if ( ($_) and ($conf{"VAR:$_"}) );
    }
    
    ## Run the command for the alert specified (save it's textual output)
    printmsg("Running the command [$alert{'commandTmp'}]", 1);
    $textOutput = `$alert{'commandTmp'}`;
    $exitStatus = $?;
    if ($exitStatus) { $exitStatus = ($exitStatus / 256); }
    $textOutput =~ s/\r|\n/ /sg;
    
    
    
    
    ################################
    ##  Log the command's output  ##
    ################################
    if ($alert{'command-log'}) {
        
        ## Make sure our file exists, and if the file doesn't exist then create it
        if ( ! -f $alert{'command-log'} ) {
            printmsg("NOTICE: The file [$alert{'command-log'}] does not exist.  Creating it now with mode [0600].", 0);
            chmod (0600, $alert{'command-log'}) unless ($^O =~ /win/i);
        }
        
        ## Now open the file and attach it to a filehandle
        if (! open (FILE,">>$alert{'command-log'}")) {
            printmsg("OS-ERROR - Opening the file [$alert{'command-log'}] for appending returned the error: $!", 0);
        }
        else {
            ## Now print the output of the last command to this file
            printmsg("OS-DEBUG - Writing the STDOUT output of the last system command to [$alert{'command-log'}]", 1);
            print FILE "$textOutput\n";
            close FILE;
        }
    }
    
    
    
    
    
    ##########################################
    ##  Evaluate the status of the command  ##
    ##########################################
    
    ## success-type => exit status
    if ($alert{'success-type'} eq "exit status") {
        printmsg("OS-DEBUG - The exit status of [$conf{'alert'}] was [$exitStatus], which means since the success-type is \"exit status\" the command has executed successfully.", 1);
        if ($exitStatus == 0) {
            quit("OS-SUCCESS - The alert command [$conf{'alert'}] executed successfully", 0);
        }
        else {
            printmsg("OS-WARNING - The alert command [$conf{'alert'}] exited with a non-zero error status of [$exitStatus].  It's textual output (if any) was [$textOutput].", 0);
        }
    }
    
    
    ## success-type => inverse exit status
    elsif ($alert{'success-type'} eq "inverse exit status") {
        printmsg("OS-DEBUG - The exit status of [$conf{'alert'}] was [$exitStatus], which means since the success-type is \"inverse exit status\" the command has failed.", 1);
        if ($exitStatus != 0) {
            quit("OS-SUCCESS - The alert command [$conf{'alert'}] executed successfully", 0);
        }
        else {
            printmsg("OS-WARNING - The alert command [$conf{'alert'}] exited with a zero error status of [$exitStatus].  It's textual output (if any) was [$textOutput].", 0);
        }
    }
    
    
    ## success-type => text
    elsif ($alert{'success-type'} eq "text") {
        if ($textOutput =~ /$alert{'success-text'}/is) {
            quit("OS-SUCCESS - The alert command [$conf{'alert'}] executed successfully", 0);
        }
        else {
            printmsg("OS-WARNING - The output of the alert command [$conf{'alert'}] did not match the \"success-text\" of [$alert{'success-text'}].  It's textual output (if any) was [$textOutput].", 0);
        }
    }
    
    
    ## success-type => undefined (generate error)
    else {
        quit("OS-ALERT - Error evaluating the status of the command for the alert [$conf{'alert'}] because an invalid \"success-type\" of [$alert{'success-type'}] was specified.", 1);
    }
    
    
    
    
    
    #################################
    ##  The command failed, so...  ##
    #################################
    
    ## Increment the global number of failed alerts
    $conf{'alertFailures'}++;
    
    ## Run the failure-alert if we need to
    if ( ($alert{'failure-alert'}) && (fork() == 0) ) {
        printmsg("OS-DEBUG - Forked process to run failure-alert [$alert{'failure-alert'}]", 1);
        $conf{'message'} = "OS-WARNING - The alert command [$conf{'alert'}] failed with an exit status of [$exitStatus].  The command's textual output (if any) was [$textOutput].  The original message (if any) was [" . $conf{'message'} . "]";
        $conf{'alert'} = $alert{'failure-alert'};
        $conf{'retriesCounter'} = 0;
    }
    else {
        
        
        ## If we need to keep retrying the current alert
        if ( ($alert{'retries'}) and ($conf{'retriesCounter'} < $alert{'retries'}) ) {
            printmsg("OS-DEBUG - We need to keep retrying the current alert ($conf{'retriesCounter'} < $alert{'retries'})", 1);
            
            ## Increment our retry counter
            $conf{'retriesCounter'}++;
            
            ## Sleep if we need to
            if ($alert{'sleep'}) {
                printmsg("OS-DEBUG - Sleeping [$alert{'sleep'}] seconds until our next retry", 1);
                sleep($alert{'sleep'});
            }
        }
        
        
        
        ## If we need to switch to the fallback-alert
        elsif ( ($alert{'fallback-alert'}) and ( (!$alert{'retries'}) or ($conf{'retriesCounter'} >= $alert{'retries'}) ) ) {
            printmsg("OS-DEBUG - Switching to the fallback-alert [$alert{'fallback-alert'}]", 1);
            $conf{'alert'} = $alert{'fallback-alert'};
            $conf{'retriesCounter'} = 0;
        }
        
        ## Otherwise there is not fallback command, so we just exit.
        else {
            quit("OS-WARNING - alertManager exiting without successfully executing the alert [$conf{'alert'}].  The original message (if any) was [$conf{'message'}]", 1);
        }
        
    }
    
}    
## Quit (should never get here)
quit("",0);
































###############################################################################################
## Function:    help ()
##
## Description: For all those newbies ;) 
##              Prints a help message and exits the program.
## 
###############################################################################################
sub help {
print <<EOM;

$conf{'programName'}-$conf{'version'} by $conf{'authorName'} <$conf{'authorEmail'}>

Usage:  $conf{'programName'} <action> [options]
  
  Actions:
    -a <alert>                    alert to run (alerts defined in the configuration file)
    --list                        list available alerts in the configuration file
    --validate                    check the syntax of the configuration file
    
  Options:
    -m <message>                  message text - normally used with "-a <alert>"
    -c <file>,--config=<file>     configuration file [$conf{'configurationFile'}]
    -v                            verbosity - use multiple times for greater effect
  
EOM
exit(1);
}










###############################################################################################
##  Function: initialize ()
##  
##  Does all the script startup jibberish.
##  
###############################################################################################
sub initialize {

  ## Set STDOUT to flush immediatly after each print  
  $| = 1;

  ## Intercept signals
  $SIG{'QUIT'}  = sub { quit("EXITING: Received SIG$_[0]", 1); };
  $SIG{'INT'}   = sub { quit("EXITING: Received SIG$_[0]", 1); };
  $SIG{'KILL'}  = sub { quit("EXITING: Received SIG$_[0]", 1); };
  $SIG{'TERM'}  = sub { quit("EXITING: Received SIG$_[0]", 1); };
  
  ## ALARM and HUP signals are not supported in Win32
  unless ($^O =~ /win/i) {
      $SIG{'HUP'}   = sub { quit("EXITING: Received SIG$_[0]", 1); };
      $SIG{'ALRM'}  = sub { quit("EXITING: Received SIG$_[0]", 1); };
  }
  
  return(1);
}









###############################################################################################
##  Function: processCommandLine ()
##  
##  Processes command line storing important data in global var %conf
##  
###############################################################################################
sub processCommandLine {
    
    
    ############################
    ##  Process command line  ##
    ############################
    
    my @ARGS = @ARGV;
    my $numargv = @ARGS;
    help() unless ($numargv);
    my $counter = 0;
    for ($counter = 0; $counter < $numargv; $counter++) {
  
        if ($ARGS[$counter] =~ /^-h$|^--help$/i) {           ## Help ##
            help();
        }
        
        elsif ($ARGS[$counter] =~ s/^--config=//i) {         ## Config File ##
            $conf{'configurationFile'} = $';
        }
        
        elsif ($ARGS[$counter] =~ /^-c$/i) {                 ## Config File ##
            $counter++;
            $conf{'configurationFile'} = $ARGS[$counter];
        }
        
        elsif ($ARGS[$counter] =~ /^-a$/i) {                 ## Alert ##
            $counter++;
            $conf{'alert'} = $ARGS[$counter];
            $conf{'mode'} = "alert";
        }
        
        elsif ($ARGS[$counter] =~ /^-m$/i) {                 ## Message ##
            $counter++;
            $conf{'message'} = $ARGS[$counter];
        }
        
        elsif ($ARGS[$counter] =~ /^--list$/i) {             ## List Alerts ##
            $conf{'mode'} = "list";
        }
        
        elsif ($ARGS[$counter] =~ /^--validate$/i) {         ## Validate Configuration ##
            $conf{'mode'} = "validate";
        }
        
        elsif ($ARGS[$counter] =~ s/^-v+//i) {               ## Verbosity ##
            my $tmp = (length($&) - 1);
            $conf{'debug'} += $tmp;
        }
        
        elsif ($ARGS[$counter] =~ /^-h$|^--help$/i) {        ## Help ##
            help();
        }
        
        else {
            printmsg("Error: \"$ARGS[$counter]\" is not a recognised option!", 0);
            help();
        }
        
    }
    
    
    ###################################################
    ##  Verify required variables are set correctly  ##
    ###################################################
    if ($conf{'mode'} !~ /alert|list|validate/i) {
        quit("OS-ERROR - One of the following command line paraters must be used: -a, --list, --validate", 1);
    }
    
    if ($conf{'mode'} eq "alert") {
        if (!$conf{'alert'}) { quit("OS-ERROR - If you are running an alert you must specify a message", 1); }
    }
    
    my @required = (
                      'configurationFile',
                      'mode',
    );
    foreach (@required) {
        if (!$conf{$_}) {
            quit("ERROR: Value [$_] was not set after parsing command line arguments!", 1);
        }
    }
    
    ## Return 0 errors
    return(0);
}



















###############################################################################################
## FUNCTION:    
##   readConfigurationFile ( string $filename, string $section, reference $hash_reference )
## 
## 
## DESCRIPTION: 
##   Reads the file $filename, and stores information from the file into the hash 
##   at $hash_reference.  This version of this function is specifically modified to support 
##   alert.pl's style configuration
##
## Example: 
##   readConfigurationFile("/etc/configuration.conf", "General Configuation", \%myHash);
##
###############################################################################################
sub readConfigurationFile {
    
    ## Get incoming variables
    my %incoming = ();
    (
      $incoming{'file'},
      $incoming{'section'},
      $incoming{'hashref'},
    ) = @_;
    
    
    ## Verify input
    if ( (!$incoming{'file'}) or (!$incoming{'section'}) or (!$incoming{'hashref'}) ) {
        printmsg("readConfigurationFile() - Incoming parameters are: $incoming{'file'} : $incoming{'section'} : $incoming{'hashref'}", 1);
        $conf{'error'} = "readConfigurationFile - OS-ERROR - One or more incoming parameters were empty";
        return(1);
    }
    
    ## Open the file
    open(CONFFILE, $incoming{'file'}) or quit("OS-ERROR - While opening the configuration file [$incoming{'file'}].  The error was [$!]\n",1);
    
    ## Read the file into a single scalar variable
    my $file = "";
    while (<CONFFILE>) {
        $file .= $_;
    }
    close(CONFFILE);
    
    ## Remove comments and whitespace
    $file =~ s/\/\*.*?\*\///ogs;      ## Remove C Style /* */ style comments
    $file =~ s/\#.*$//ogm;            ## Remove pound (#) comments
    $file =~ s/\/\/.*$//ogm;          ## Remove slash (//) comments
    $file =~ s/^\s*(\n|\r\n)//ogm;    ## Remove blank lines
    
    ## Get the section that we're looking for
    if (! ($file =~ s/$incoming{'section'}\s*\{.*?\}//s) ) {
        $conf{'error'} = "ERROR - The definition [$incoming{'section'}] could not be found in the configuration file!";
        return(2);
    }
    
    ## Remove the header and any left-over whitespace
    $file = $&;
    $file =~ s/(^$incoming{'section'}\s*\{)|(\}$)//sg;
    
    ## For each line in that section store the data
    for my $line (split(/\n|\r\n/, $file)) {
        ## Remove whitespace
        $line =~ s/(^\s*)|(\s*$)//og;
        ## Don't continue unless the line looks valid
        next if ( (!$line) or ($line !~ /=>/) );
        ${$incoming{'hashref'}}{(split(/\s*=>\s*/, $line))[0]} = (split(/\s*=>\s*/, $line))[1] unless (${$incoming{'hashref'}}{(split(/\s*=>\s*/, $line))[0]});    
    }
    
    ## Return 0 errors
    return(0);
}


















###############################################################################################
## FUNCTION:    
##   returnConfigurationFileSections ( string $filename )
## 
## 
## DESCRIPTION: 
##   Reads the file $filename, finds any defined sections, and returns a list of them.  This
##   is designed for configuration files that use alert.pl's style configuration.
##
## Example: 
##   my @list = returnConfigurationFileSections("/etc/configuration.conf");
##
###############################################################################################
sub returnConfigurationFileSections {
    
    ## Get incoming variables
    my %incoming = ();
    (
      $incoming{'file'},
    ) = @_;
    
    ## Verify input
    if ( (!$incoming{'file'}) ) {
        $conf{'error'} = "returnConfigurationFileSections - OS-ERROR - One or more incoming parameters were empty";
        return(1);
    }
    
    ## Open the file
    open(CONFFILE, $incoming{'file'}) or quit("OS-ERROR While opening the configuration file [$incoming{'file'}].  The error was [$!]\n",1);
    
    
    ## Read the file into a single scalar variable
    my $file = "";
    while (<CONFFILE>) {
        $file .= $_;
    }
    close(CONFFILE);
    
    ## Remove comments and whitespace
    $file =~ s/\/\*.*?\*\///ogs;      ## Remove C Style /* */ style comments
    $file =~ s/\#.*$//ogm;            ## Remove pound (#) comments
    $file =~ s/\/\/.*$//ogm;          ## Remove slash (//) comments
    $file =~ s/^\s*(\n|\r\n)//ogm;    ## Remove blank lines
    
    ## Get each section one at a time
    my @entries = ();
    while ($file =~ s/(.*?\s*\{.*?\}\s*)//is) {
        my $entry = $&;
        $entry =~ s/\s*\{.*?\}\s*//s;
        push(@entries, $entry);
    }
    
    ## Check for duplicate entries
    foreach (@entries) {
        my $entry = $_;
        if (scalar(grep(/^$entry$/, @entries)) >= 2) {
            printmsg("OS-WARNING - There are two alert definitions in the configuration file [$incoming{'file'}] named [$entry]", 0);
        }
    }
    
    ## Return 0 errors
    return(@entries);
}


















###############################################################################################
## FUNCTION:    
##   validateConfigurationFile ( string $filename )
## 
## 
## DESCRIPTION: 
##   This function will only be useful for AlertManager!
##   Reads the file $filename, and does dependancy checking, and validation on all data in 
##   the config file.  It returns 0 on success and non-zero on configuration error.  Configuartion
##   error messages will be printed via the printmsg function, errors generated elsewhere in the
##   program will usually be set in $conf{'error'}, so you will want to print it if there's any-
##   thing in it.
##
##   
###############################################################################################
sub validateConfigurationFile {
    
    ## Get incoming variables
    my %incoming = ();
    (
      $incoming{'file'},
    ) = @_;
    
    ## Verify input
    if ( (!$incoming{'file'}) ) {
        $conf{'error'} = "validateConfigurationFile - OS-ERROR - One or more incoming parameters were empty";
        return(1);
    }
    
    ## Get a list of sections in the configuration file
    my @list = returnConfigurationFileSections($incoming{'file'});
    @list = grep(!/General Settings/, @list);
    
    ## Start an error counter
    $conf{'errors'} = 0;
    
    ## Start the validation process
    foreach my $entryName (@list) {
        printmsg("validateConfigurationFile() - OS-DEBUG - Processing entry [$entryName]", 2);
        my %entry = ();
        
        ## Read the current section into the array %entry
        readConfigurationFile($conf{'configurationFile'}, $entryName, \%entry) && do {
            return(2);
        };
        
        ## Now verify each type of data they could have stored
        for my $key (keys(%entry)) {
            
            ## command
            if ($key eq "command") {
                if (!$entry{$key}) { printmsg(qq{ERROR - In the alert definition [$entryName], the "command" value was empty}, 0); $conf{'errors'}++; next; }
            }
            
            ## command-log
            elsif ($key eq "command-log") {
                if (!$entry{$key}) { printmsg(qq{ERROR - In the alert definition [$entryName], the "command-log" value was empty}, 0); $conf{'errors'}++; next; }
            }
            
            
            ## success-type
            elsif ($key eq "success-type") {
                if (!$entry{$key}) { printmsg(qq{ERROR - In the alert definition [$entryName], the "success-type" value was empty}, 0); $conf{'errors'}++; next; }
                if ($entry{$key} !~ /text|exit status|inverse exit status/) { printmsg(qq{ERROR - In the alert definition [$entryName], the "success-type" value must be either "text", "exit status", or "inverse exit status"}, 0); $conf{'errors'}++; }
            }
            
            
            ## success-text
            elsif ($key eq "success-text") {
                if (!$entry{$key}) { printmsg(qq{ERROR - In the alert definition [$entryName], the "success-text" value was empty}, 0); $conf{'errors'}++; next; }
                if ($entry{'success-type'} ne "text") { printmsg(qq{WARNING - In the alert definition [$entryName], the "success-text" value will be ignored because the "success-type" was not "text"}, 0); $conf{'errors'}++; }
            }
            
            
            ## retries
            elsif ($key eq "retries") {
                if (!$entry{$key}) { printmsg(qq{ERROR - In the alert definition [$entryName], the "retries" value was empty}, 0); $conf{'errors'}++; next; }
                if ($entry{$key} =~ /\D/) { printmsg(qq{ERROR - In the alert definition [$entryName], the "retries" value must be numeric}, 0); $conf{'errors'}++; }
                if ($entry{$key} > 99)    { printmsg(qq{ERROR - In the alert definition [$entryName], the "retries" value must not be greater than 99}, 0); $conf{'errors'}++; }
            }
            
            
            ## sleep
            elsif ($key eq "sleep") {
                if (!$entry{$key}) { printmsg(qq{ERROR - In the alert definition [$entryName], the "sleep" value was empty}, 0); $conf{'errors'}++; next; }
                if ($entry{$key} =~ /\D/) { printmsg(qq{ERROR - In the alert definition [$entryName], the "sleep" value must be numeric}, 0); $conf{'errors'}++; }
                if ($entry{$key} > 999)   { printmsg(qq{ERROR - In the alert definition [$entryName], the "retries" value must not be greater than 999}, 0); $conf{'errors'}++; }
            }
            
            
            ## failure-alert
            elsif ($key eq "failure-alert") {
                if (!$entry{$key}) { printmsg(qq{ERROR - In the alert definition [$entryName], the "failure-alert" value was empty}, 0); $conf{'errors'}++; next; }
                unless ($entry{$key} =~ /^LOG\s+.+|^SYSLOG$/) {
                    my %tmp = ();
                    readConfigurationFile($conf{'configurationFile'}, $entry{$key}, \%tmp) && do { 
                        printmsg(qq{ERROR - In the alert definition [$entryName], the specified "failure-alert" of "$entry{$key}" couldn't be loaded from the configuration file}, 0);
                        $conf{'errors'}++;
                    };
                }
            }
            
            
            ## fallback-alert
            elsif ($key eq "fallback-alert") {
                if (!$entry{$key}) { printmsg(qq{ERROR - In the alert definition [$entryName], the "fallback-alert" value was empty}, 0); $conf{'errors'}++; next; }
                unless ($entry{$key} =~ /^LOG\s+.+|^SYSLOG$/) {
                    my %tmp = ();
                    readConfigurationFile($conf{'configurationFile'}, $entry{$key}, \%tmp) && do { 
                        printmsg(qq{ERROR - In the alert definition [$entryName], the specified "fallback-alert" of "$entry{$key}" couldn't be loaded from the configuration file}, 0);
                        $conf{'errors'}++;
                    };
                }
            }
            
            
            ## An unknown option
            else {
                printmsg(qq{ERROR - In the alert definition [$entryName], I don't understand the option [$key].}, 0);
                $conf{'errors'}++;
            }
            
            
        }
        
        
        
        
    }
    
    ## Return number of errors
    return($conf{'errors'});
}


















###############################################################################################
# FUNCTION: handleHUP()
# 
# Handles HUP signals:
#  - Re-reads the configuration file
#  - Re-parses the command line
#
###############################################################################################
sub handleHUP  {
    
    printmsg("Reloading configuration", 1);
    
    
    #####################################################
    ##  Reset some variables that should be refreshed  ##
    #####################################################
    $conf{'logFile'} = '';
    $conf{'debug'} = 0;
    
    
    ######################################
    ##  Re-read the configuration file  ##
    ######################################
    readConfigurationFile($conf{'configurationFile'}, "General Settings", \%conf) && do {
        quit("$conf{'error'}", 1);
    };
    
    ## Get list of user defined custom variables (if any) - this is used later in the script
    @userVars = grep(/^VAR:/, keys(%conf));
    for (my $a = 0; $a < scalar(@userVars); $a++) { 
        $userVars[$a] =~ s/^VAR://;
    }
    
    #######################################################
    ##  Re-parse the command line - it takes precedence  ##
    #######################################################
    processCommandLine();
    
    
    ## Debugging: Show the current configuration
    printmsg ("Configuration: ", 3); foreach my $key (keys(%conf))           { printmsg ("Key: $key =>\t Value: $conf{$key}", 3) if ($key and $conf{$key}); }
    printmsg ("User defined variables: ", 3);   foreach my $key (@userVars)         { printmsg ("Key: $key", 3) if ($key); }
    
    ## Go back to the normal routine
    return(1);
}














###############################################################################################
##  Function:    printmsg (string $message, int $level)
##
##  Description: Handles all messages - logging them to a log file, 
##               printing them to the screen or both depending on
##               the $level passed in, $conf{'debug'} and wether
##               $conf{'mode'}.
##
##  Input:       $message          A message to be printed, logged, etc.
##               $level            The debug level of the message. If
##                                 not defined 0 will be assumed.  0 is
##                                 considered a normal message, 1 and 
##                                 higher is considered a debug message.
##  
##  Output:      Prints to STDOUT, to LOGFILE, both, or none depending 
##               on the state of the program.
##  
##  Example:     ("WARNING: We believe in generic error messages... NOT!\n", 1);
###############################################################################################
sub printmsg {
    my %incoming = ();
    (
      $incoming{'message'},
      $incoming{'level'}
    ) = @_;
    $incoming{'level'} = 0 if (!defined($incoming{'level'}));
    $incoming{'message'} =~ s/\r|\n/ /sg;
    $incoming{'message'} =~ s/\s*$//sg;
    
    ## Continue on if the debug level is >= the incoming message level
    if ($conf{'debug'} >= $incoming{'level'}) {
        
        ## Print to STDOUT always
        print localtime() . " - $$ - $conf{'programName'} - $incoming{'message'}\n";
        
        ## Print to the log file if a logFile is defined
        if ( ($conf{'mode'} eq "alert") and ($conf{'logFile'}) ) {
            if (openLogFile($conf{'logFile'})) { $conf{'logFile'} = ""; printmsg("OS-ERROR - Opening the file [$conf{'logFile'}] for appending returned the error: $!", 1); }
            else {
                print LOGFILE localtime() . " - $$ - $conf{'programName'} - $incoming{'message'}\n";
                close LOGFILE;
            }
        }
    
    }
    
    ## Return 0 errors
    return(0);
}












###############################################################################################
## FUNCTION:    
##   openLogFile ( $filename )
## 
## 
## DESCRIPTION: 
##   Opens the file $filename and attaches it to the filehandle "LOGFILE".  Returns 0 on success
##   and non-zero on failure.  Error codes are listed below, and the error message gets set in
##   global variable $!.
##   
##   
## Example: 
##   openFile ("/var/log/scanAlert.log");
##
###############################################################################################
sub openLogFile {
    ## Get the incoming filename
    my $filename = $_[0];
    
    ## Make sure our file exists, and if the file doesn't exist then create it
    if ( ! -f $filename ) {
        print STDERR "NOTICE: The log file [$filename] does not exist.  Creating it now with mode [0600].\n";
        open (LOGFILE, ">>$filename");
        close LOGFILE;
        chmod (0600, $filename);
    }
    
    ## Now open the file and attach it to a filehandle
    open (LOGFILE,">>$filename") or return (1);
    
    ## Put the file into non-buffering mode
    select LOGFILE;
    $| = 1;
    select STDOUT;
    
    ## Tell the rest of the program that we can log now
    $conf{'logging'} = "yes";
    
    ## Return success
    return(0);
}











###############################################################################################
## FUNCTION:    
##   syslog (string $socketName, int $priority, int $facility, string $message)
## 
## 
## DESCRIPTION: 
##   Connects to the socket $socketName, and sends it a syslog formatted message.  
##   i.e. it syslog's $message with the priority and facility specified.
##
## Example: 
##   syslog("/dev/log", 6, 1, "Test syslog message");
##
##
## Priorities (on a Linux system)
##   LOG_EMERG       0       system is unusable
##   LOG_ALERT       1       action must be taken immediately
##   LOG_CRIT        2       critical conditions
##   LOG_ERR         3       error conditions
##   LOG_WARNING     4       warning conditions
##   LOG_NOTICE      5       normal but significant condition
##   LOG_INFO        6       informational
##   LOG_DEBUG       7       debug-level messages
##
##
## Facilities (on a Linux system)
##   LOG_KERN        0       kernel messages
##   LOG_USER        1       random user-level messages
##   LOG_MAIL        2       mail system
##   LOG_DAEMON      3       system daemons
##   LOG_AUTH        4       security/authorization messages
##   LOG_SYSLOG      5       messages generated internally by syslogd
##   LOG_LPR         6       line printer subsystem
##   LOG_NEWS        7       network news subsystem
##   LOG_UUCP        8       UUCP subsystem
##   LOG_CRON        9       clock daemon
##   LOG_AUTHPRIV    10      security/authorization messages (private)
##   LOG_FTP         11      ftp daemon
##
###############################################################################################
sub syslog {
    
    ## Get incoming variables
    my %incoming = ();
    (
      $incoming{'socketName'},
      $incoming{'priority'},
      $incoming{'facility'},
      $incoming{'message'},
    ) = @_;
    
    ## Set defaults if some variables are not defined
    $incoming{'socketName'}  = "/dev/log"         if (!defined($incoming{'socketName'}));
    $incoming{'priority'}    = 6                  if (!defined($incoming{'priority'}));
    $incoming{'facility'}    = 1                  if (!defined($incoming{'facility'}));
    $incoming{'message'}     = "Default Message"  if (!defined($incoming{'message'}));
  
    ## Make sure values are sane (black magic)
    $incoming{'priority'} = $incoming{'priority'} & 0x07;
    $incoming{'facility'} = $incoming{'facility'} & 0x3f8;
    
    
    ## Debuging
    printmsg("syslog(): Syslogging the message [$incoming{'message'}]", 1);
    
    if (! -S $incoming{'socketName'}) {
        $conf{'error'} = "OS-ERROR - The specified syslog socket [$incoming{'socketName'}] either does not exist or is not a socket.";
        return(1);
    }
    
    ## Open a UNIX socket in dgram mode with udp protocol. 
    socket(SOCKET, PF_UNIX, SOCK_DGRAM, 0) or do { 
        $conf{'error'} = "syslog() - OS-ERROR - $!";
        return(2);
    };
    
    ## Connect our socket to SOCKET
    connect(SOCKET, sockaddr_un ($incoming{'socketName'})) or do {
        $conf{'error'} = "syslog() - OS-ERROR - $!";
        return(3);
    };
    
    ## Generate syslog value for the priority and facility (black magic)
    $incoming{'value'} = ($incoming{'priority'} + (8 * $incoming{'facility'}));
  
    ## Sending the message to the syslog socket
    print SOCKET "<$incoming{'value'}>$incoming{'message'}" or do {
        $conf{'error'} = "syslog() - OS-ERROR - $!";
        return(4);
    };
    
    ## Close the socket
    close SOCKET;
    
    
    ## Return success
    return(0);
}











###############################################################################################
##  Function:    quit (string $message, int $errorLevel)
##  
##  Description: Exits the program, optionally printing $message.  It 
##               returns an exit error level of $errorLevel to the 
##               system  (0 means no errors, and is assumed if empty.)
##
##  Example:     quit("Exiting program normally", 0);
###############################################################################################
sub quit {
    my %incoming = ();
    (
      $incoming{'message'},
      $incoming{'errorLevel'}
    ) = @_;
    $incoming{'errorLevel'} = 0 if (!defined($incoming{'errorLevel'}));
    
    ## Print exit message
    if ($incoming{'message'}) { 
        printmsg($incoming{'message'}, 0);
        
#        ## Syslog the message if exiting with an error status
#        if ( ($conf{'debug'} == 0) and (-S "/dev/log") ) {
#            syslog("/dev/log", 6, 1, $incoming{'message'});
#        }
        
    }
    
    ## Exit
    exit($incoming{'errorLevel'});
}






